import os
import tarfile
import netCDF4
import numpy as np


def read(archive):
    path, basename = os.path.split(archive)
    _, ext = os.path.splitext(basename)

    prod = {}

    def _handle_fid(fid, symbol="Z"):
        if symbol == "Z":
            with netCDF4.Dataset("memory", mode="r", memory=fid.read()) as nc:
                atts = nc.ncattrs()
                name = nc.getncattr("TypeName")
                prod.update({"longitude": nc.getncattr("Longitude")})
                prod.update({"latitude": nc.getncattr("Latitude")})
                prod.update({"sweep_el": nc.getncattr("Elevation")})
                prod.update({"sweep_az": nc.getncattr("Azimuth")})
                prod.update({"time": nc.getncattr("Time")})
                prod.update({"prf": float(round(nc.getncattr("PRF-value") * 0.1) * 10.0)})
                prod.update({"waveform": nc.getncattr("Waveform") if "Waveform" in atts else ""})
                prod.update({"gatewidth": float(nc.variables["GateWidth"][:][0])})
                prod.update({"elevations": np.array(nc.variables["Elevation"][:], dtype=np.float32)})
                prod.update({"azimuths": np.array(nc.variables["Azimuth"][:], dtype=np.float32)})
                prod.update({symbol: np.array(nc.variables[name][:], dtype=np.float32)})
        else:
            with netCDF4.Dataset("memory", mode="r", memory=fid.read()) as nc:
                name = nc.getncattr("TypeName")
                prod.update({symbol: np.array(nc.variables[name][:], dtype=np.float32)})

    # Read from a tar archive
    if ext in ['.tar', '.tgz', '.xz']:
        prod.update({"archive": archive})
        with tarfile.open(archive) as tar:
            for member in tar.getmembers():
                if not member.isfile():
                    continue
                file, _ = os.path.splitext(os.path.basename(member.name))
                parts = file.split("-")
                symbol = parts[4]
                with tar.extractfile(member) as fid:
                    _handle_fid(fid, symbol)
    # Read from netCDF file
    elif ext == ".nc":
        symbols = ["Z", "V", "W", "D", "P", "R"]
        parts = basename.split("-")
        for symbol in symbols:
            parts[-1] = f"{symbol}.nc"
            basename = "-".join(parts)
            filename = os.path.join(path, basename)
            print(f"filename: {filename}")
            with open(filename, mode="rb") as fid:
                _handle_fid(fid, symbol)
    else:
        print(f"Unknown file extension: {ext}")
        return None

    # Generate coordinate arrays that are 1 element extra than each dimension
    de = prod["elevations"][-1] - prod["elevations"][-2]
    prod["e"] = [*prod["elevations"], prod["elevations"][-1] + de]
    prod["r"] = np.arange(prod["Z"].shape[1] + 1) * 1.0e-3 * prod["gatewidth"]
    symbols = ["Z", "V", "W", "D", "P", "R"]
    mask = prod["Z"] < -999.0
    for key in symbols:
        prod[key][mask] = np.nan
    return prod



from . import sweep
from . import overlay

import numpy as np
import matplotlib.patheffects
import matplotlib.pyplot as plt

import blib


def rho2ind(values):
    m3 = values > 0.93
    m2 = np.logical_and(values > 0.7, ~m3)
    index = values * 52.8751
    index[m2] = values[m2] * 300.0 - 173.0
    index[m3] = values[m3] * 1000.0 - 824.0
    return index


def shade(shape, xy=(0.5, 0.5), rgba=[0, 0, 0, 0.5], direction="southeast"):
    if direction == "n" or direction == "north":
        x = np.ones((shape[1],))
        y = blib.colormap.fleximap(shape[0], [0, xy[1], 1], [[1], [1], [0]])
        a = np.outer(y, x)
    elif direction == "ne" or direction == "northeast":
        x = blib.colormap.fleximap(shape[1], [0, xy[0], 1], [[0], [1], [1]])
        y = blib.colormap.fleximap(shape[0], [0, xy[1], 1], [[1], [1], [0]])
        a = np.outer(y, x)
    elif direction == "e" or direction == "east":
        x = blib.colormap.fleximap(shape[1], [0, xy[0], 1], [[0], [1], [1]])
        y = np.ones((shape[0],))
        a = np.outer(y, x)
    elif direction == "se" or direction == "southeast":
        x = blib.colormap.fleximap(shape[1], [0, xy[0], 1], [[0], [1], [1]])
        y = blib.colormap.fleximap(shape[0], [0, xy[1], 1], [[0], [1], [1]])
        a = np.outer(y, x)
    elif direction == "s" or direction == "south":
        x = np.ones((shape[1],))
        y = blib.colormap.fleximap(shape[0], [0, xy[1], 1], [[0], [1], [1]])
        a = np.outer(y, x)
    elif direction == "sw" or direction == "southwest":
        x = blib.colormap.fleximap(shape[1], [0, xy[0], 1], [[1], [1], [0]])
        y = blib.colormap.fleximap(shape[0], [0, xy[1], 1], [[0], [1], [1]])
        a = np.outer(y, x)
    elif direction == "w" or direction == "west":
        x = blib.colormap.fleximap(shape[1], [0, xy[0], 1], [[1], [1], [0]])
        y = np.ones((shape[0],))
        a = np.outer(y, x)
    elif direction == "nw" or direction == "northwest":
        x = blib.colormap.fleximap(shape[1], [0, xy[0], 1], [[1], [1], [0]])
        y = blib.colormap.fleximap(shape[0], [0, xy[1], 1], [[1], [1], [0]])
        a = np.outer(y, x)
    else:
        x = blib.colormap.fleximap(shape[1], [0, xy[0], 1], [[0], [1], [0]])
        y = blib.colormap.fleximap(shape[0], [0, xy[1], 1], [[0], [1], [0]])
        a = np.outer(y, x)
    m = np.empty((*shape, 4))
    m[:, :, :] = rgba
    m[:, :, 3] *= a
    return m


def ribbon():
    z = blib.colormap.fleximap(
        20,
        [0.0, 0.50, 0.51, 1.0],
        [[0.33, 0.60, 0.69, 0.70], [0.12, 0.39, 0.48, 0.70], [0.00, 0.27, 0.36, 0.70], [0.00, 0.27, 0.36, 0.70]],
    )
    z = z.reshape((20, 1, 4))
    return z


class Chart:
    size = (1280, 720)
    seed = 111
    dpi = 72
    s = 1.0
    frameon = True
    titlecolor = None
    orientation = "horizontal"
    fig = None
    # elevation, azimuth, range
    e = None
    a = None
    r = None
    title = None
    overlay = None
    axis_is_set = False
    labelfont = blib.getFontOfWeight(weight=600)
    titlefont = blib.getFontOfWeight(weight=700)
    symbols = []

    def __init__(self, n=1, **kwargs):
        """
        Create a new chart.

        Optional Parameters
        -------------------
        size: (float, float), default: (1280, 720)
            The size of the chart in pixels.

        dpi: float, default: 72
            The resolution of the chart in dots per inch.

        s: float, default: 1.0
            The scaling factor for the chart.

        titlecolor: (float, float, float), default: :rc:`text.color`
            The color of the title.

        orientation: str, default: "horizontal"
            The orientation of the colorbars.

        Returns
        -------
        A new chart.
        """
        if "size" in kwargs:
            self.size = kwargs["size"]
        if "seed" in kwargs:
            self.seed = kwargs["seed"]
        if "dpi" in kwargs:
            self.dpi = kwargs["dpi"]
        if "s" in kwargs:
            self.s = kwargs["s"]
        if "titlecolor" in kwargs:
            self.titlecolor = kwargs["titlecolor"]
        if "orientation" in kwargs:
            self.orientation = kwargs["orientation"]
        self.m = 5 * self.s
        self.p = 10 * self.s
        self.labelsize = 12 * self.s
        self.titlesize = 28 * self.s
        self.captionsize = 16 * self.s
        self.ax = [None] * n
        self.cb = [None] * n
        self.st = [None] * n
        self.ms = [None] * n

        width, height = self.size
        self.figsize = (width / self.dpi, height / self.dpi)

        color = matplotlib.rcParams["text.color"]
        if color in matplotlib.colors.CSS4_COLORS:
            rgba = matplotlib.colors.to_rgba(matplotlib.colors.CSS4_COLORS[color])
            hsv = matplotlib.colors.rgb_to_hsv(rgba[:3])
        elif color[0] == "#":
            rgba = matplotlib.colors.to_rgba(color)
            hsv = matplotlib.colors.rgb_to_hsv(rgba[:3])
        else:
            hsv = (0, 0, 1)
        if hsv[2] > 0.5:
            self.gridcolor = (0.2, 1.0, 1.0)
            self.titlecolor = (0.8, 0.8, 0.8) if self.titlecolor is None else self.titlecolor
        else:
            self.gridcolor = (0, 0.4, 0.4)
            self.titlecolor = (0.2, 0.2, 0.2) if self.titlecolor is None else self.titlecolor
        self.path_effects = [
            matplotlib.patheffects.Stroke(linewidth=2 * self.s, foreground=matplotlib.rcParams["axes.facecolor"]),
            matplotlib.patheffects.Normal(),
        ]
        self.figprops = {
            "axes.linewidth": max(1, self.s // 1),
            "axes.axisbelow": False,
            "axes.facecolor": kwargs.get("facecolor", (1, 1, 1, 0.85) if hsv[2] < 0.5 else (0, 0, 0, 0.85)),
            "axes.titlepad": 10 * self.s,
            "axes.labelpad": 6 * self.s,
            "axes.linewidth": max(1, self.s // 1),
            "xtick.direction": "in",
            "xtick.major.pad": 8 * self.s,
            "xtick.major.size": round(3 * self.s),
            "xtick.major.width": max(1, self.s // 1),
            "xtick.labelsize": self.labelsize,
            "ytick.direction": "in",
            "ytick.major.pad": 8 * self.s,
            "ytick.major.size": round(3 * self.s),
            "ytick.major.width": max(1, self.s // 1),
            "ytick.labelsize": self.labelsize,
        }

    def _get_pos(self, num):
        # Available space for axes
        width, height = self.size
        ww = width
        hh = height - self.titlesize - self.m
        if num > 1000:
            cols = num // 1000
            rows = (num - cols * 1000) // 100
            i = num % 100
        else:
            cols = num // 100
            rows = (num - cols * 100) // 10
            i = num % 10
        w = round((ww - (cols + 1) * self.m) / cols) / width
        h = round((hh - (rows + 1) * self.m) / rows) / height
        if cols < rows:
            x = (i - 1) // rows
            y = rows - 1 - (i - 1) % rows
        else:
            x = (i - 1) % cols
            y = rows - 1 - (i - 1) // cols
        # print(f'x = {x}  y = {y}  w = {w}  h = {h}')
        x = (x * w) + (x + 1) * self.m / width
        y = (y * h) + (y + 1) * self.m / height
        return [x, y, w, h]

    def _draw_line(self, q, c="#f7931a"):
        """
        Draw a line on the chart.

        Parameters
        ----------
        q : (float, float, float, float)
            (origin_x, origin_y, width, height)
            The position of the line.
        c : str, color default: "#f7931a"
        """
        x0 = q[0]
        y0 = q[1]
        x1 = q[0] + q[2]
        y1 = q[1] + q[3]
        x = [x0, x1]
        y = [y0, y1]
        line = matplotlib.lines.Line2D(x, y, color=c, linewidth=self.s)
        self.fig.add_artist(line)

    def _draw_box(self, q, xoff=0, yoff=0, c=None, b="#4d4d4d"):
        """
        Draw a box on the chart.

        Parameters
        ----------
        q : (float, float, float, float)
            (origin_x, origin_y, width, height)
            The position of the box.
        xoff : float, default: 0
            The x offset of the box.
        yoff : float, default: 0
            The y offset of the box.
        c : str, color, default: None
        b : str, background color, default: "#4d4d4d"
        """
        width, height = self.size
        x0 = q[0]
        x1 = x0 + 10 / width
        x2 = q[0] + q[2]
        y0 = q[1]
        y1 = q[1] + q[3]
        x0 += xoff / width
        y0 += yoff / height
        x = [x1, x2, x2, x0, x0, x1]
        y = [y0, y0, y1, y1, y0, y0]
        if c is not None:
            line = matplotlib.lines.Line2D(x, y, color=c, linewidth=self.s, solid_joinstyle="miter")
            self.fig.add_artist(line)
        if b is not None:
            rect = matplotlib.patches.Rectangle(q[0:2], q[2], q[3], color=b, zorder=-1)
            self.fig.add_artist(rect)

    def _add_axes(self, num):
        width, height = self.size
        q = self._get_pos(num)
        ax = self.fig.add_axes(q, frameon=False, snap=True, xticks=[], yticks=[])
        if self.frameon:
            self._draw_box(q, c=matplotlib.rcParams["text.color"], b=matplotlib.rcParams["axes.facecolor"])
        else:
            self._draw_box(q, c="#f7931a")
        t = 10 * self.s
        p = self.p
        shade_color = matplotlib.colors.to_rgb(matplotlib.rcParams["axes.facecolor"])

        if self.orientation[:4] == "vert":
            # Background shade, from right: p, 3 * labelsize, p, t, p, captionsize, 2p
            w = 5 * p + 3 * self.labelsize + t + self.captionsize
            bq = [
                q[0] + q[2] - w / width,
                q[1],
                w / width,
                q[3],
            ]
            z = shade((80, 50), (0.4, 0.5), [*shade_color, 0.5], direction="se")
            # Colorbar
            c = min(q[3] * height - 4 * p, 512 * self.s) / height
            w = 3 * self.captionsize + t
            cq = [
                q[0] + q[2] - w / width,
                q[1] + 2 * p / height,
                t / width,
                c,
            ]
        else:
            # Background shade, from top: p, captionsize, p, t, p, labelsize, 2p
            h = 5 * p + self.captionsize + t + self.labelsize
            bq = [
                q[0],
                q[1] + q[3] - h / height,
                q[2],
                h / height,
            ]
            z = shade((50, 80), (0.4, 0.5), [*shade_color, 0.5], direction="nw")
            # Colorbar
            c = min(q[2] * width - 4 * p, 512 * self.s) / width
            h = 2 * p + self.captionsize + t
            cq = [
                q[0] + 2 * p / width,
                q[1] + q[3] - h / height,
                c,
                t / height,
            ]
        # Axis for background shade
        bx = self.fig.add_axes(bq, frameon=False, snap=True)
        bx.imshow(z, aspect="auto")
        bx.set(xticks=[], yticks=[])
        # Axis for colorbar
        cb = self.fig.add_axes(cq, frameon=False, snap=True)
        cb.set(xticks=[], yticks=[])
        self._draw_box(cq, xoff=-1, c=matplotlib.rcParams["text.color"], b=matplotlib.rcParams["axes.facecolor"])
        if self.orientation[:4] == "vert":
            st = matplotlib.text.Text(cq[0] - (self.captionsize + p) / width, cq[1] + 0.5 * cq[3], f"{num}")
            st.set(
                fontproperties=self.labelfont,
                horizontalalignment="center",
                path_effects=self.path_effects,
                rotation=self.orientation,
                size=self.captionsize,
                verticalalignment="center",
            )
            self.fig.add_artist(st)
        else:
            st = cb.set_title(f"{num}", fontproperties=self.labelfont)
            st.set(size=self.captionsize, path_effects=self.path_effects)
        return ax, cb, st

    def _update_data_only(self, sweep: sweep.Sweep, **kwargs):
        if "symbol" in kwargs:
            symbols = list([kwargs["symbol"]])
        else:
            symbols = kwargs.get("symbols", sweep.products.keys())
        self.symbols = list(symbols)
        for m, symbol in zip(self.ms, self.symbols[: len(self.ax)]):
            if symbol == "R":
                m.set_array(rho2ind(sweep.products[symbol]).ravel())
                continue
            m.set_array(sweep.products[symbol].ravel())
        if "title" in kwargs:
            self.update_title(kwargs.get("title"))

    def _update_coordinate_data(self, xx, yy, sweep: sweep.Sweep, **kwargs):
        if self.overlay and all([m is not None for m in self.ms]):
            for m in self.ms:
                if m is None:
                    continue
                m.remove()
            self.ms = [None] * len(self.ax)
        # Update self.symbols
        if "symbol" in kwargs:
            symbols = list([kwargs["symbol"]])
        else:
            symbols = kwargs.get("symbols", sweep.products.keys())
        self.symbols = list(symbols)
        props = {
            "shading": "flat",
            "zorder": 1,
        }
        for k, (ax, symbol) in enumerate(zip(self.ax, self.symbols)):
            if "map" in kwargs:
                value = kwargs["map"](sweep.products[symbol])
                cmap = kwargs.get("cmap", blib.matplotlibColormap("zmapx"))
                vmin = kwargs.get("vmin", np.min(value.flatten()))
                vmax = kwargs.get("vmax", np.max(value.flatten()))
            elif symbol[0] == "Z":
                value = sweep.products[symbol]
                cmap = blib.matplotlibColormap("rsz")
                vmin, vmax = -32, 96
            elif symbol[0] == "V":
                value = sweep.products[symbol]
                cmap = blib.matplotlibColormap("v")
                vmin, vmax = -64, 64
            elif symbol[0] == "W":
                value = sweep.products[symbol]
                cmap = blib.matplotlibColormap("w")
                vmin, vmax = 0, 12.8
            elif symbol == "D":
                value = sweep.products[symbol]
                cmap = blib.matplotlibColormap("rsd")
                vmin, vmax = -10, 15.5
            elif symbol == "P":
                value = sweep.products[symbol]
                cmap = blib.matplotlibColormap("p")
                vmin, vmax = -180, 180
            elif symbol == "R":
                value = rho2ind(sweep.products[symbol])
                cmap = blib.matplotlibColormap("rsr")
                vmin, vmax = 0, 256
            elif symbol == "RR":
                value = np.log10(sweep.products[symbol])
                cmap = blib.matplotlibColormap("rr")
                vmin, vmax = -1, 2
            else:
                value = sweep.products[symbol]
                cmap = kwargs.get("cmap", blib.matplotlibColormap("zmapx"))
                vmin = kwargs.get("vmin", np.min(value.flatten()))
                vmax = kwargs.get("vmax", np.max(value.flatten()))
            self.ms[k] = ax.pcolormesh(xx, yy, value, cmap=cmap, vmin=vmin, vmax=vmax, **props)

        if sweep.scanType == "rhi":
            if "xmax" in kwargs:
                self.set_xlim(0, kwargs["xmax"])
            if "ymax" in kwargs:
                self.set_ylim(0, kwargs["ymax"])
        else:
            if "xmax" in kwargs:
                xmax = kwargs["xmax"]
                self.set_xlim(-xmax, xmax)
            if "ymax" in kwargs:
                ymax = kwargs["ymax"]
                self.set_ylim(-ymax, ymax)
            if "xmax" not in kwargs and "ymax" not in kwargs and self.axis_is_set is False:
                rmax = kwargs.get("rmax", np.hypot(np.max(np.abs(xx)), np.max(np.abs(yy))))
                xoff = kwargs.get("xoff", 0.0)
                yoff = kwargs.get("yoff", 0.0)
                aspect = self.ax[0].bbox.width / self.ax[0].bbox.height
                if aspect > 1:
                    self.set_xlim(-rmax * aspect + xoff, rmax * aspect + xoff)
                    self.set_ylim(-rmax + yoff, rmax + yoff)
                else:
                    self.set_xlim(-rmax + xoff, rmax + xoff)
                    self.set_ylim(-rmax / aspect + yoff, rmax / aspect + yoff)
                self.axis_is_set = True
        self.r = sweep.meshCoordinate.r
        self.a = sweep.meshCoordinate.a
        self.e = sweep.meshCoordinate.e

    def _setup_colorbars(self, **kwargs):
        # Colorbars
        for m, a, c in zip(self.ms, self.ax, self.cb):
            plt.colorbar(m, cax=c, ax=a, orientation=self.orientation)
        # Colorbar labels
        for k, symbol in enumerate(self.symbols[: len(self.ax)]):
            if symbol[0] == "Z":
                self.st[k].set_text(f"{symbol} - Reflectivity (dBZ)")
            elif symbol[0] == "V":
                self.st[k].set_text(f"{symbol} - Velocity (m/s)")
            elif symbol[0] == "W":
                self.st[k].set_text(f"{symbol} - Spectrum Width (m/s)")
            elif symbol == "P":
                self.st[k].set_text("P - Differential Phase (°)")
            elif symbol == "D":
                self.st[k].set_text("D - Differential Reflectivity (dB)")
            elif symbol == "R":
                self.st[k].set_text("R - Correlation Coefficient")
            elif symbol == "RR":
                self.st[k].set_text("RR - Rain Rate (mm/hr)")

        # Colorbar ticks
        tick_props = {
            "fontproperties": self.labelfont,
            "fontsize": self.labelsize,
            "path_effects": self.path_effects,
        }

        def setup_ticks(k, ticks, ticklabels, lo, hi):
            if self.orientation == "vertical":
                self.cb[k].set_yticks(ticks, labels=ticklabels, **tick_props)
                self.cb[k].set_ylim(lo, hi)
            else:
                self.cb[k].set_xticks(ticks, labels=ticklabels, **tick_props)
                self.cb[k].set_xlim(lo, hi)

        for k, symbol in enumerate(self.symbols[: len(self.ax)]):
            if "map" in kwargs:
                ticks = kwargs.get("ticks", np.linspace(kwargs["vmin"], kwargs["vmax"], 5))
                setup_ticks(k, kwargs["map"](ticks), ticks, kwargs["vmin"], kwargs["vmax"])
                continue
            elif symbol[0] == "Z":
                ticks = np.arange(-20, 61, 20)
                setup_ticks(k, ticks, ticks, -10, 75)
            elif symbol[0] == "V":
                ticks = np.arange(-20, 21, 10)
                setup_ticks(k, ticks, ticks, -30, 30)
            elif symbol[0] == "W":
                ticks = np.arange(2, 9, 2)
                setup_ticks(k, ticks, ticks, 0, 10)
            elif symbol == "P":
                ticks = np.arange(-120, 121, 60)
                setup_ticks(k, ticks, ticks, -180, 180)
            elif symbol == "D":
                ticks = np.arange(-4, 5, 2)
                setup_ticks(k, ticks, ticks, -6, 6)
            elif symbol == "R":
                ticks = np.array([0.73, 0.83, 0.93, 0.96, 0.99, 1.02])
                setup_ticks(k, rho2ind(ticks), ticks, 0, 180)
            elif symbol == "RR":
                ticks = np.array([0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50])
                setup_ticks(k, np.log10(ticks), ticks, -1, 2)

    def update_title(self, text):
        if self.title is None:
            return self.set_title(text)
        if self.title.get_horizontalalignment() != "left" or len(self.title.get_text()) != len(text):
            extent = self.title.get_window_extent()
            self.title.remove()
            # Do a test run to get the proper position of the title, then use left alignment to avoid jitters in animations
            y = (self.size[1] - self.titlesize - self.m) / self.size[1]
            title_props = {
                "fontproperties": self.titlefont,
                "fontsize": self.titlesize,
                "color": self.titlecolor,
                "horizontalalignment": "center",
                "verticalalignment": "bottom",
            }
            title = self.fig.text(0.5, y, text, **title_props)
            extent = title.get_window_extent()
            title.remove()
            title_props["horizontalalignment"] = "left"
            x = extent.x0 / self.size[0] * self.dpi / self.fig.dpi
            self.title = self.fig.text(x, y, text, **title_props)
            return
        self.title.set_text(text)

    def set_title(self, text):
        if self.title:
            if len(self.title.get_text()) == len(text):
                self.title.set_text(text)
                return
            self.title.remove()
        y = (self.size[1] - self.titlesize - self.m) / self.size[1]
        title_props = {
            "fontproperties": self.titlefont,
            "fontsize": self.titlesize,
            "color": self.titlecolor,
            "horizontalalignment": "center",
            "verticalalignment": "bottom",
        }
        self.title = self.fig.text(0.5, y, text, **title_props)

    def set_data(self, sweep: sweep.Sweep, **kwargs):
        if kwargs.get("verbose", 0) > 1:
            print("Chart.set_data() called with kwargs:", kwargs)
        """
        Set the data for the chart.

        Optional Parameters
        -------------------
        title: str
        rmax: float
        xoff: float
        yoff: float
        """
        if self.r is sweep.meshCoordinate.r and self.e is sweep.meshCoordinate.e and "symbol" not in kwargs:
            return self._update_data_only(sweep, **kwargs)

        if sweep.scanType == "rhi":

            e_rad = np.radians(sweep.meshCoordinate.e)
            xx = np.outer(np.cos(e_rad), 1.0e-3 * sweep.meshCoordinate.r)
            yy = np.outer(np.sin(e_rad), 1.0e-3 * sweep.meshCoordinate.r)
            self._update_coordinate_data(xx, yy, sweep, **kwargs)

            if self.overlay is None:
                # Find out the meaningful range of the plot
                mask = np.logical_and(sweep.products["Z"] > -20, sweep.products["Z"] < 80)
                h_max = kwargs.get("ymax", np.ceil(np.max(yy[1:, 1:][mask]) + 3.5))
                extent = (0, 0, 1.0e-3 * sweep.meshCoordinate.r[-1], max(sweep.meshCoordinate.e))
                self.overlay = overlay.PolarGrid(extent=extent, ymax=h_max, s=self.s)
                self.overlay.load()
                for ax in self.ax:
                    self.overlay.draw(ax)

        elif sweep.scanType == "ppi":

            e_rad = np.radians(sweep.sweepElevation)
            a_rad = np.radians(sweep.meshCoordinate.a)
            xx = np.outer(np.cos(e_rad) * np.sin(a_rad), 1.0e-3 * sweep.meshCoordinate.r)
            yy = np.outer(np.cos(e_rad) * np.cos(a_rad), 1.0e-3 * sweep.meshCoordinate.r)
            self._update_coordinate_data(xx, yy, sweep, **kwargs)

            if self.overlay is None:
                origin = (sweep.longitude, sweep.latitude)
                xlim, ylim = self.ax[0].get_xlim(), self.ax[0].get_ylim()
                extent = (xlim[0], ylim[0], xlim[1], ylim[1])
                density = self.ax[0].bbox.width / (xlim[1] - xlim[0])
                self.overlay = overlay.Overlay(origin=origin, extent=extent, density=density, **kwargs)
                self.overlay.load()
                if self.orientation == "vertical":
                    exclude = (0.75 * xlim[1] + 0.25 * xlim[0], xlim[1], ylim[0], 0.25 * ylim[0] + 0.75 * ylim[1])
                    self.overlay.exclude(exclude)
                for ax in self.ax:
                    self.overlay.draw(ax)
        else:

            raise ValueError("Unknown scan type.")

        self._setup_colorbars(**kwargs)
        if sweep.time is None and "title" not in kwargs:
            return
        self.update_title(kwargs.get("title", sweep.time.strftime(r"%Y/%m/%d %H:%M:%S UTC")))
        if "colorbar_title" in kwargs:
            self.st[0].set_text(kwargs["colorbar_title"])

    def set_xlim(self, lo, hi=None):
        for ax in self.ax:
            ax.set_xlim(lo, hi)

    def set_ylim(self, lo, hi=None):
        for ax in self.ax:
            ax.set_ylim(lo, hi)

    def set(self, **kwargs):
        if "title" in kwargs:
            self.set_title(kwargs["title"])
        if "xlim" in kwargs:
            self.set_xlim(kwargs["xlim"][0], kwargs["xlim"][1])
        if "ylim" in kwargs:
            self.set_ylim(kwargs["ylim"][0], kwargs["ylim"][1])

    def close(self):
        plt.close(self.fig)


class ChartRHI(Chart):
    seed = 231

    def __init__(self, sweep: sweep.Sweep = None, **kwargs):
        """
        Create a new RHI chart.

        Parameters
        ----------
        sweep : sweep.Sweep
            The sweep to be displayed.

        Returns
        -------
        A new RHI chart.
        """
        super().__init__(n=6, **kwargs)

        with plt.rc_context(self.figprops):
            self.fig = plt.figure(figsize=self.figsize, dpi=self.dpi, frameon=False)
            for i in range(6):
                self.ax[i], self.cb[i], self.st[i] = self._add_axes(self.seed + i)

        if sweep:
            self.set_data(sweep, **kwargs)


class ChartPPI(Chart):
    seed = 321

    def __init__(self, sweep: sweep.Sweep = None, **kwargs):
        """
        Create a new PPI chart.

        Parameters
        ----------
        sweep : sweep.Sweep
            The sweep to be displayed.

        Returns
        -------
        A new PPI chart.
        """
        if "orientation" not in kwargs:
            kwargs["orientation"] = "vertical"
        super().__init__(n=6, **kwargs)

        with plt.rc_context(self.figprops):
            self.fig = plt.figure(figsize=self.figsize, dpi=self.dpi, frameon=False)
            for i in range(6):
                self.ax[i], self.cb[i], self.st[i] = self._add_axes(self.seed + i)

        if sweep:
            self.set_data(sweep, **kwargs)

    def set(self, **kwargs):
        if "rmax" in kwargs:
            aspect = self.ax[0].bbox.width / self.ax[0].bbox.height
            rmax = kwargs["rmax"]
            if aspect < 1:
                for ax in self.ax:
                    ax.set(xlim=(-rmax, rmax), ylim=(-rmax * aspect, rmax * aspect))
            else:
                for ax in self.ax:
                    ax.set(xlim=(-rmax * aspect, rmax * aspect), ylim=(-rmax, rmax))

    def set_rmax(self, rmax):
        self.set(rmax=rmax)

    # TODO: Add methods to set rmax, x-offset, y-offset, etc.


class ChartRHITall(ChartRHI):
    seed = 161
    size = (640, 1440)


class ChartRHIWide(ChartRHI):
    seed = 611
    size = (3840, 250)


class ChartSinglePPI(Chart):
    size = (1920, 1080)
    seed = 111
    s = 1.25
    frameon = False

    def __init__(self, sweep: sweep.Sweep = None, **kwargs):
        """
        Create a new Single PPI chart.

        Parameters
        ----------
        sweep : sweep.Sweep
            The sweep to be displayed.

        Returns
        -------
        A new PPI chart.
        """
        super().__init__(n=1, **kwargs)

        # Overrides
        self.titlesize = 36 * self.s
        self.path_effects = [
            matplotlib.patheffects.Stroke(linewidth=self.s, foreground=matplotlib.rcParams["axes.facecolor"]),
            matplotlib.patheffects.Normal(),
        ]

        def get_colorbar_rect():
            ch = round(16.0 * self.s)  # Colorbar height
            pw = round(25.0 * self.s)  # Colorbar padding width
            ph = round(12.0 * self.s)  # Colorbar padding height
            cw = round(512 * self.s)  # Colorbar width
            # Reserve 100 pts for the big symbol
            while cw > self.size[0] - self.s * 100:
                cw -= 128 if self.size[0] < 640 else 256
            rect = [
                (self.size[0] - cw - pw) / self.size[0],
                (self.size[1] - ch - ph - self.captionsize) / self.size[1],
                cw / self.size[0],
                ch / self.size[1],
            ]
            return rect

        with plt.rc_context(self.figprops):
            self.fig = plt.figure(figsize=self.figsize, dpi=self.dpi, frameon=False)
            self.ax[0] = self.fig.add_axes([0, 0, 1, 1], frameon=False, snap=True, xticks=[], yticks=[])
            self._draw_box([0, 0, 1, 1], b="black")

            # Axis for background shade
            h = 70 * self.s / self.size[1]
            bx = self.fig.add_axes([0, 1.0 - h, 1, h], frameon=False, snap=True, xticks=[], yticks=[])
            bx.imshow(ribbon(), aspect="auto")
            self._draw_line([0, 1.0 - h + 1.5 / self.size[1], 1, 0], c=[1, 1, 1, 0.5])

            # Colorbar
            cq = get_colorbar_rect()
            self.cb[0] = self.fig.add_axes(cq, frameon=True, snap=True, xticks=[], yticks=[])
            self._draw_box(cq, xoff=-1, c=matplotlib.rcParams["text.color"], b=matplotlib.rcParams["axes.facecolor"])
            colorbar_title_props = {
                "fontproperties": self.labelfont,
                "fontsize": self.captionsize,
                "pad": 5,
                "verticalalignment": "bottom",
                "path_effects": self.path_effects,
            }
            self.st[0] = self.cb[0].set_title(f"{1}", **colorbar_title_props)

            # Title
            title_props = {
                "fontproperties": self.titlefont,
                "fontsize": self.titlesize,
                "color": self.titlecolor,
                "horizontalalignment": "left",
                "verticalalignment": "top",
                "path_effects": self.path_effects,
            }
            x = 25.0 * self.s / self.size[0]
            y = 1.0 - 0.58 * (h - self.titlesize / self.size[1])
            self.title = self.fig.text(x, y, "", **title_props)

        if sweep:
            self.set_data(sweep, **kwargs)



import os
import re
import radar
import datetime
import numpy as np

from ._ctypes_ import RKProductCollectionInitWithFilename

re_3parts = re.compile(
    r"(?P<name>.+)-"
    + r"(?P<time>20[0-9][0-9](0[0-9]|1[012])([0-2][0-9]|3[01])-([01][0-9]|2[0-3])[0-5][0-9][0-5][0-9])-"
    + r"(?P<scan>[EAN][0-9]+\.[0-9])"
)


def wrapDegrees(angle):
    return (angle + 180.0) % 360.0 - 180.0


class MeshCoordinate:
    e = [0]
    a = [0]
    r = [0]
    x = [0]
    y = [0]
    z = [0]
    kind = "polar"

    def __repr__(self):
        if self.kind == "cartesian":
            return f"MeshCoordinate({len(self.z)} x {len(self.y)} x {len(self.x)})"
        elif self.kind == "polar":
            return f"MeshCoordinate({len(self.e)} x {len(self.a)} x {len(self.r)})"


symbolMapping = {"DBZ": "Z", "VEL": "V", "WIDTH": "W", "ZDR": "D", "PHIDP": "P", "RHOHV": "R"}


class Sweep:
    archive = None
    prefix = None
    time = None
    longitude = -97.0
    latitude = 32.0
    scanType = "unk"
    sweepElevation = 0.0
    sweepAzimuth = 0.0
    prf = 1
    waveform = "u"
    gatewidth = 1.0
    elevations = 0.0
    azimuths = 0.0
    products = {}
    meshCoordinate = MeshCoordinate()

    def __init__(self, source=None, **kwargs):
        """
        Sweep class to handle radar data

        Parameters
        ----------
        input : str or np.ndarray
            If str, it is the path to a file or a tar archive containing radar data.
            Currently support RadarKit format only.

            If np.ndarray, it is a 3D array of radar data with shape (6, n, m) where
            n is the number of elevation or azimuth angles and m is the number of range gates.

        Returns
        -------
        A Sweep object
        """
        if source is None:
            return
        if isinstance(source, np.ndarray) and source.ndim == 3 and (source.shape[0] == 6 or "symbols" in kwargs):
            self.plain(source, **kwargs)
        elif isinstance(source, str) and os.path.exists(source):
            self.file(source)

    def __repr__(self):
        a = (
            self.sweepElevation
            if self.scanType.lower() == "ppi"
            else self.sweepAzimuth if self.scanType.lower() == "rhi" else -999.0
        )
        return f"Sweep: {self.scanType} {a:.1f} - {list(self.products.keys())}"

    def _get_prefix(self):
        if self.archive is None:
            return
        basename = os.path.basename(self.archive)
        if "_V06" in basename:
            self.prefix = basename[:4]
        else:
            self.prefix = basename.split("-")[0]

    def file(self, filename):
        collection = RKProductCollectionInitWithFilename(filename).contents
        product = collection.products[0]
        self.archive = filename
        self._get_prefix()
        self.time = datetime.datetime.fromtimestamp(product.header.startTime, tz=datetime.timezone.utc)
        self.longitude = product.header.longitude
        self.latitude = product.header.latitude
        self.scanType = "ppi" if product.header.isPPI else "rhi" if product.header.isRHI else "unk"
        self.sweepElevation = product.header.sweepElevation
        self.sweepAzimuth = product.header.sweepAzimuth
        self.prf = 1.0 / product.header.prt[0]
        self.waveform = str(product.header.waveformName)
        self.gatewidth = product.header.gateSizeMeters
        self.elevations = np.ctypeslib.as_array(product.startElevation, (product.header.rayCount,))
        self.azimuths = np.ctypeslib.as_array(product.startAzimuth, (product.header.rayCount,))
        self.products = {}
        for k in range(collection.count):
            product = collection.products[k]
            symbol = str(product.desc.symbol.decode("utf-8"))
            symbol = symbolMapping.get(symbol, symbol)
            values = np.ctypeslib.as_array(product.data, (product.header.rayCount, product.header.gateCount))
            self.products.update({symbol: values})
        # Generate coordinate arrays that are 1 element extra than each dimension
        if self.scanType == "rhi":
            de = self.elevations[-1] - self.elevations[-2]
            self.meshCoordinate.e = [*self.elevations, self.elevations[-1] + de]
        elif self.scanType == "ppi":
            da = self.azimuths[-1] - self.azimuths[-2]
            self.meshCoordinate.a = [*self.azimuths, self.azimuths[-1] + da]
        self.meshCoordinate.r = np.arange(self.products["Z"].shape[1] + 1) * self.gatewidth

    def plain(self, array, **kwargs):
        symbols = kwargs.get("symbols", ["Z", "V", "W", "D", "P", "R"])
        self.products = {}
        for i, symbol in enumerate(symbols):
            self.products.update({symbol: array[i, :, :]})
        self.scanType = kwargs["scanType"] if "scanType" in kwargs else kwargs.get("type", "unk")
        self.archive = kwargs.get("archive", None)
        self._get_prefix()
        if "time" in kwargs:
            self.time = kwargs["time"]
            if isinstance(self.time, float) or isinstance(self.time, int):
                self.time = datetime.datetime.fromtimestamp(self.time, tz=datetime.timezone.utc)
        if "latitude" in kwargs:
            self.latitude = kwargs["latitude"]
        if "longitude" in kwargs:
            self.longitude = kwargs["longitude"]
        if "gatewidth" in kwargs:
            self.gatewidth = kwargs["gatewidth"]
        elif "dr" in kwargs:
            self.gatewidth = kwargs["dr"]
        if "e" in kwargs:
            self.meshCoordinate.e = kwargs["e"]
            if len(self.meshCoordinate.e) == array.shape[1]:
                de = self.meshCoordinate.e[-1] - self.meshCoordinate.e[-2]
                self.meshCoordinate.e = np.append(self.meshCoordinate.e, self.meshCoordinate.e[-1] + de)
        if "a" in kwargs:
            self.meshCoordinate.a = kwargs["a"]
            if len(self.meshCoordinate.a) == array.shape[1]:
                da = self.meshCoordinate.a[-1] - self.meshCoordinate.a[-2]
                self.meshCoordinate.a = np.append(self.meshCoordinate.a, wrapDegrees(self.meshCoordinate.a[-1] + da))
        if "r" in kwargs:
            self.meshCoordinate.r = kwargs["r"]
            if len(self.meshCoordinate.r) == array.shape[2]:
                dr = self.meshCoordinate.r[-1] - self.meshCoordinate.r[-2]
                self.meshCoordinate.r = np.append(self.meshCoordinate.r, self.meshCoordinate.r[-1] + dr)
        if "x" in kwargs:
            self.meshCoordinate.kind = "cartesian"
            self.meshCoordinate.x = kwargs["x"]
            if len(self.meshCoordinate.x) == array.shape[2]:
                dx = self.meshCoordinate.x[-1] - self.meshCoordinate.x[-2]
                self.meshCoordinate.x = np.append(self.meshCoordinate.x, self.meshCoordinate.x[-1] + dx)
        if "y" in kwargs:
            self.meshCoordinate.kind = "cartesian"
            self.meshCoordinate.y = kwargs["y"]
            if len(self.meshCoordinate.y) == array.shape[1]:
                dy = self.meshCoordinate.y[-1] - self.meshCoordinate.y[-2]
                self.meshCoordinate.y = np.append(self.meshCoordinate.y, self.meshCoordinate.y[-1] + dy)
        # Fix the meshCoordinate arrays to have one extra element in each dimension
        if self.meshCoordinate.kind == "polar":
            if len(self.meshCoordinate.e) != array.shape[1] + 1:
                self.meshCoordinate.e = np.arange(array.shape[1] + 1, dtype=float) * kwargs.get("de", 1.0)
            if len(self.meshCoordinate.a) != array.shape[1] + 1:
                self.meshCoordinate.a = wrapDegrees(np.arange(array.shape[1] + 1, dtype=float) * kwargs.get("da", 1.0))
            if len(self.meshCoordinate.r) != array.shape[2] + 1:
                self.meshCoordinate.r = np.arange(array.shape[2] + 1, dtype=float) * self.gatewidth
        elif self.meshCoordinate.kind == "cartesian":
            if len(self.meshCoordinate.x) != array.shape[2] + 1:
                self.meshCoordinate.x = np.arange(array.shape[2] + 1, dtype=float) * kwargs.get("dx", 1.0)
            if len(self.meshCoordinate.y) != array.shape[1] + 1:
                self.meshCoordinate.y = np.arange(array.shape[1] + 1, dtype=float) * kwargs.get("dy", 1.0)
        # TODO: Think about z
        self.elevations = self.meshCoordinate.e[:-1]
        self.azimuths = self.meshCoordinate.a[:-1]
        self.range = self.meshCoordinate.r[:-1]
        self.sweepElevation = kwargs.get("sweepElevation", np.mean(self.elevations))
        self.sweepAzimuth = kwargs.get("sweepAzimuth", np.mean(self.azimuths))
        if self.scanType == "unk":
            ib = int(0.2 * len(self.azimuths))
            ie = int(0.8 * len(self.azimuths))
            if ie - ib > 10:
                if self.elevations[ib:ie].std() < 1.0:
                    self.scanType = "ppi"
                elif self.azimuths[ib:ie].std() < 1.0:
                    self.scanType = "rhi"



import os
import blib
import json
import matplotlib
import matplotlib.patheffects
import numpy as np

prefix = os.path.join(os.path.split(__file__)[0], "maps")
r_earth = 6378.0
pop_big = 50000
pop_med = 95000  # Norman is 95694


class MapColor:
    def __init__(self, theme="light"):
        if theme == "light":
            self.ring = "#0099cc"
            self.state = "#40bf91"
            self.county = "#40bf91"
            self.highway = "#e6b955"
            self.text = "#000000"
        else:
            self.ring = "#78dcff"
            self.state = "#96e6c8"
            self.county = "#83e2bf"
            self.highway = "#e6b955"
            self.text = "#ffffff"


colors = MapColor("dark")

if not os.path.exists(prefix):
    prefix, _ = os.path.split(__file__)
    prefix = os.path.join(prefix, "maps")
    if not os.path.exists(prefix):
        raise FileNotFoundError(f"Could not find maps directory: {prefix}")


def _read(name="county"):
    if name == "state":
        filename = "states-10m.stq.json"
    elif name == "county":
        filename = "gz_2010_us_050_00_500k.stq.json"
    elif name == "interstate" or name == "highway":
        filename = "intrstat.stq.json"
    elif name == "city":
        filename = "citiesx020.shp.json"
    else:
        filename = name
    path = os.path.join(prefix, filename)
    if not os.path.exists(path):
        print(f"Could not find map: {path}")
        return None
    with open(path, "r") as fid:
        data = json.load(fid)
    return data


def makeRotationX(phi):
    return np.array([[1.0, 0.0, 0.0], [0.0, np.cos(phi), np.sin(phi)], [0.0, -np.sin(phi), np.cos(phi)]])


def makeRotationY(phi):
    return np.array([[np.cos(phi), 0.0, -np.sin(phi)], [0.0, 1.0, 0.0], [np.sin(phi), 0.0, np.cos(phi)]])


def makeRotationZ(phi):
    return np.array([[np.cos(phi), np.sin(phi), 0.0], [-np.sin(phi), np.cos(phi), 0.0], [0.0, 0.0, 1.0]])


def makeRotationForCoord(lon=-97.46381, lat=35.23682):
    return np.matmul(makeRotationY(np.radians(-lon)), makeRotationX(np.radians(lat)))


def project(coords, rotation=makeRotationForCoord()):
    lon, lat = coords[:, 0], coords[:, 1]
    m = r_earth * np.cos(lat)
    y = r_earth * np.sin(lat)
    z = m * np.cos(lon)
    x = m * np.sin(lon)
    p = np.array((x, y, z)).transpose()
    return np.matmul(p, rotation)


def coordsFromPolyNaive(poly):
    coords = []
    w = poly["transform"]["scale"]
    b = poly["transform"]["translate"]
    for arc in poly["arcs"]:
        lat = w[1] * arc[0][1] + b[1]
        if lat < -89:
            continue
        line = []
        point = [0, 0]
        for p in arc:
            point[0] += p[0]
            point[1] += p[1]
            lon = w[0] * point[0] + b[0]
            lat = w[1] * point[1] + b[1]
            line.append([lon, lat])
        coords.append(np.array(line))
    return coords


def coordsFromPoly(poly):
    coords = []
    w = poly["transform"]["scale"]
    b = poly["transform"]["translate"]
    for arc in poly["arcs"]:
        lat = w[1] * arc[0][1] + b[1]
        if lat < -89:
            continue
        line = np.cumsum(np.array(arc), axis=0)
        coords.append(line * w + b)
    return coords


def get(poly, origin=(-97.46381, 35.23682), extent=(-160, -90, 160, 90), density=6.0):
    if isinstance(poly, str):
        if poly[:5] == "@ring":
            r = np.array([float(s) for s in poly.split("/")[1:]])
            a = np.arange(0, 2.01 * np.pi, np.pi / 180)
            x, y = np.cos(a), np.sin(a)
            rings = []
            for k in r:
                rings.append(k * np.array([x, y]).transpose())
            return rings
        poly = _read(name=poly)
        return get(poly, origin=origin, extent=extent, density=density)
    if isinstance(poly, dict) and "transform" in poly:
        coords = coordsFromPoly(poly)
        x_min, y_min, x_max, y_max = extent
        rotation = makeRotationForCoord(*origin)
        subset = []
        for c in coords:
            pos = project(np.radians(c), rotation)
            inside = np.logical_and(
                np.logical_and(pos[:, 0] > x_min, pos[:, 0] < x_max),
                np.logical_and(pos[:, 1] > y_min, pos[:, 1] < y_max),
            )
            if np.any(inside):
                subset.append(pos[:, :2])
        return subset
    elif isinstance(poly, list) and isinstance(poly[0], dict) and "G" in poly[0]:
        coords = np.array([t["G"]["C"] for t in poly])
        labels = [t["P"]["N"] for t in poly]
        pops = [t["P"]["P"] for t in poly]
        x_min, y_min, x_max, y_max = extent
        rotation = makeRotationForCoord(*origin)
        pos = project(np.radians(coords), rotation)
        rr = np.hypot(pos[:, 0], pos[:, 1])
        inside = np.logical_and(
            np.logical_and(pos[:, 0] > x_min, pos[:, 0] < x_max), np.logical_and(pos[:, 1] > y_min, pos[:, 1] < y_max)
        )
        popped = [
            2500 * np.log(p) / p < density if p > 0 else r < 100 if density > 4 else r < 10 or density > 99
            for p, r in zip(pops, rr)
        ]
        show = np.logical_and(inside, popped)
        text = [(*pos[:2], colors.text, label, pop) for pos, label, pop, s in zip(pos, labels, pops, show) if s]
        return text
    else:
        raise ValueError(f"Unknown polygon format: {type(poly)}")


def radii(max_range=70.0, about=7):
    deltas = [1, 2, 5, 10, 20, 25, 50, 100, 200, 250, 500, 1000]
    k = np.argmin(max_range / np.array(deltas) >= about)
    delta = deltas[k]
    return np.array([1.0, *np.arange(delta, max_range, delta)])


class Grid:
    c = matplotlib.colors.to_rgb(matplotlib.rcParams["axes.facecolor"])
    path_effects = [matplotlib.patheffects.Stroke(linewidth=2.5, foreground=(*c, 0.6)), matplotlib.patheffects.Normal()]
    label_props = {
        "fontproperties": blib.getFontOfWeight(weight=600),
        "horizontalalignment": "center",
        "verticalalignment": "center",
        "path_effects": path_effects,
        "zorder": 200,
    }
    s = 1
    size1 = 12
    size2 = 14
    size3 = 16
    gridwidth = 1.0
    linewidth = 1.5

    def __init__(self, **kwargs):
        c = matplotlib.rcParams["text.color"]
        if c in matplotlib.colors.CSS4_COLORS:
            rgba = matplotlib.colors.to_rgba(matplotlib.colors.CSS4_COLORS[c])
            hsv = matplotlib.colors.rgb_to_hsv(rgba[:3])
        elif c[0] == "#":
            rgba = matplotlib.colors.to_rgba(c)
            hsv = matplotlib.colors.rgb_to_hsv(rgba[:3])
        else:
            hsv = (0, 0, 1)
        global colors
        if hsv[2] > 0.5:
            colors = MapColor("dark")
        else:
            colors = MapColor("light")
        if "s" in kwargs:
            self.s = kwargs["s"]
            self.size1 = 12 * self.s
            self.size2 = 14 * self.s
            self.size3 = 16 * self.s
            c = matplotlib.colors.to_rgb(matplotlib.rcParams["axes.facecolor"])
            self.label_props["path_effects"] = [
                matplotlib.patheffects.Stroke(linewidth=2.5 * self.s, foreground=(*c, 0.6)),
                matplotlib.patheffects.Normal(),
            ]
        self.gridwidth = kwargs.get("gridwidth", max(1.0, self.s))
        self.linewidth = kwargs.get("linewidth", max(1.0, 1.5 * self.s))


"""
    extent in (xmin, ymin, xmax, ymax)
"""


class Overlay(Grid):
    labels = None
    county = None
    highway = None
    rings = None

    def __init__(self, origin=(-97.46381, 35.23682), extent=(-160, -90, 160, 90), **kwargs):
        super().__init__(**kwargs)
        self.origin = origin
        self.extent = extent
        self.density = kwargs.get("density", 4.0)
        self.radii = kwargs.get("radii", radii(max_range=np.hypot(extent[2], extent[3]), about=7))

    def _pop2size(self, pop):
        return self.size3 if pop > pop_big else self.size2 if pop > pop_med else self.size1

    def load(self):
        self.labels = get("city", self.origin, self.extent, self.density)
        self.county = get("county", self.origin, self.extent)
        self.highway = get("highway", self.origin, self.extent)
        name = "@ring/" + "/".join([f"{r:.0f}" for r in self.radii])
        self.rings = get(name, self.origin, self.extent)
        # Append ring labels to self.labels
        c, s = np.cos(np.pi / 4), np.sin(np.pi / 4)
        for r in self.radii[1:]:
            label = f"{r:.0f} km"
            self.labels.append((r * c, r * s, colors.ring, label, 1000))
            self.labels.append((-r * c, -r * s, colors.ring, label, 1000))
        self.labels.sort(key=lambda x: x[-1], reverse=True)

    def exclude(self, extent):
        removals = []
        for label in self.labels:
            x, y, _, _, _ = label
            if x < extent[0] or y < extent[1] or x > extent[2] or y > extent[3]:
                continue
            removals.append(label)
        for label in removals:
            self.labels.remove(label)

    def draw(self, ax):
        for p in self.county:
            x, y = p[:, 0], p[:, 1]
            ax.add_line(matplotlib.lines.Line2D(x, y, color=colors.county, linewidth=self.linewidth, zorder=100))
        for p in self.highway:
            x, y = p[:, 0], p[:, 1]
            ax.add_line(matplotlib.lines.Line2D(x, y, color=colors.highway, linewidth=self.linewidth, zorder=100))
        for p in self.rings:
            x, y = p[:, 0], p[:, 1]
            ax.add_line(matplotlib.lines.Line2D(x, y, color=colors.ring, linewidth=self.gridwidth, zorder=100))
        occupied = []
        for x, y, c, label, pop in self.labels:
            size = self._pop2size(pop)
            text = matplotlib.text.Text(x, y, label, color=c, fontsize=size, **self.label_props)
            ax.add_artist(text)
            rect = text.get_window_extent().padded(10)
            overlap = False
            for o in occupied:
                if o.overlaps(rect):
                    overlap = True
                    break
            if overlap:
                text.remove()
            else:
                occupied.append(rect)


"""
    extent in (rmin, emin, rmax, emax)
"""


class PolarGrid(Grid):
    grids = []
    labels = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.extent = kwargs.get("extent", (0, 0, 60, 40))
        self.ymax = kwargs.get("ymax", 20)
        self.radii = kwargs.get("radii", radii(max_range=self.extent[2], about=5))

    def load(self):
        e_max = 90 if self.extent[3] < 90 else 180
        self.grids = []
        self.ee = np.radians(np.arange(0, e_max, 2))
        xx = np.outer(np.cos(self.ee), self.radii)
        yy = np.outer(np.sin(self.ee), self.radii)

        # Range rings
        for k in range(xx.shape[1]):
            x = xx[:, k]
            y = yy[:, k]
            self.grids.append((x, y))
        # Elevation lines
        for e in np.arange(10, e_max, 10):
            e_rad = np.radians(e)
            cc = np.cos(e_rad)
            ss = np.sin(e_rad)
            x = [cc, cc * self.radii[-1]]
            y = [ss, ss * self.radii[-1]]
            self.grids.append((x, y))
        # Labels
        self.labels = []
        points = [[40.0, 10.0], [20.0, 20.0]]
        for pos in points:
            phi, r = np.radians(pos[0]), pos[1]
            x, y = r * np.cos(phi), r * np.sin(phi)
            self.labels.append((x, y, f"{r:.0f} km"))

    def draw(self, ax):
        ax.set(xlim=[0, self.extent[2]], ylim=[0, self.ymax], xticks=[], yticks=[])
        for x, y in self.grids:
            line = matplotlib.lines.Line2D(x, y, color=colors.ring, linewidth=self.gridwidth)
            ax.add_artist(line)
        for x, y, label in self.labels:
            text = matplotlib.text.Text(x, y, label, color=colors.ring, fontsize=self.size1, **self.label_props)
            ax.add_artist(text)



from ._ctypes_ import *

RKPulseStatusDict = {
    "RKPulseStatusNull": RKPulseStatusNull,
    "RKPulseStatusVacant": RKPulseStatusVacant,
    "RKPulseStatusHasIQData": RKPulseStatusHasIQData,
    "RKPulseStatusHasPosition": RKPulseStatusHasPosition,
    "RKPulseStatusInspected": RKPulseStatusInspected,
    "RKPulseStatusCompressed": RKPulseStatusCompressed,
    "RKPulseStatusSkipped": RKPulseStatusSkipped,
    "RKPulseStatusDownSampled": RKPulseStatusDownSampled,
    "RKPulseStatusProcessed": RKPulseStatusProcessed,
    "RKPulseStatusRingInspected": RKPulseStatusRingInspected,
    "RKPulseStatusRingFiltered": RKPulseStatusRingFiltered,
    "RKPulseStatusRingSkipped": RKPulseStatusRingSkipped,
    "RKPulseStatusRingProcessed": RKPulseStatusRingProcessed,
    "RKPulseStatusReadyForMomentEngine": RKPulseStatusReadyForMomentEngine,
    "RKPulseStatusCompleteForMoments": RKPulseStatusCompleteForMoments,
    "RKPulseStatusUsedForMoments": RKPulseStatusUsedForMoments,
    "RKPulseStatusProcessMask": RKPulseStatusProcessMask,
    "RKPulseStatusRecorded": RKPulseStatusRecorded,
    "RKPulseStatusStreamed": RKPulseStatusStreamed,
}

RKWaveformTypeDict = {
    "RKWaveformTypeNone": RKWaveformTypeNone,
    "RKWaveformTypeIsComplex": RKWaveformTypeIsComplex,
    "RKWaveformTypeSingleTone": RKWaveformTypeSingleTone,
    "RKWaveformTypeFrequencyHopping": RKWaveformTypeFrequencyHopping,
    "RKWaveformTypeLinearFrequencyModulation": RKWaveformTypeLinearFrequencyModulation,
    "RKWaveformTypeTimeFrequencyMultiplexing": RKWaveformTypeTimeFrequencyMultiplexing,
    "RKWaveformTypeFromFile": RKWaveformTypeFromFile,
    "RKWaveformTypeFlatAnchors": RKWaveformTypeFlatAnchors,
    "RKWaveformTypeFrequencyHoppingChirp": RKWaveformTypeFrequencyHoppingChirp,
}

RKMarkerDict = {
    "RKMarkerNull": RKMarkerNull,
    "RKMarkerSweepMiddle": RKMarkerSweepMiddle,
    "RKMarkerSweepBegin": RKMarkerSweepBegin,
    "RKMarkerSweepEnd": RKMarkerSweepEnd,
    "RKMarkerVolumeBegin": RKMarkerVolumeBegin,
    "RKMarkerVolumeEnd": RKMarkerVolumeEnd,
    "RKMarkerScanTypeMask": RKMarkerScanTypeMask,
    "RKMarkerScanTypeUnknown": RKMarkerScanTypeUnknown,
    "RKMarkerScanTypePPI": RKMarkerScanTypePPI,
    "RKMarkerScanTypeRHI": RKMarkerScanTypeRHI,
    "RKMarkerScanTytpePoint": RKMarkerScanTytpePoint,
    "RKMarkerMemoryManagement": RKMarkerMemoryManagement,
}


def show_flag(RKstatus, sdict=RKPulseStatusDict):
    s = "-"
    for iks in sdict.keys():
        if RKstatus & sdict[iks]:
            s = s + " | " + iks
    print(s)



import time
import tqdm
import ctypes
import contextlib

import numpy as np

from ._ctypes_ import *

# from .defs import *

# from . import radial_noise

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# In this module the open and close functions are overwritten, use
# __builtins__.open and __builtins__.close to do necessary development.
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

workspaces = []
Productdict = {
    "S": RKProductIndexSh,
    "Z": RKProductIndexZ,
    "V": RKProductIndexV,
    "W": RKProductIndexW,
    "D": RKProductIndexD,
    "P": RKProductIndexP,
    "R": RKProductIndexR,
}


class RKEngineError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"RKEngineError: {self.message}"


class pyRKuint32(ctypes.c_uint32):
    pass


class UserParams(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("directory", ctypes.c_char * RKMaximumFolderPathLength),
        ("filename", ctypes.c_char * RKNameLength),
        ("verbose", ctypes.c_int),
        ("output", ctypes.c_bool),
        ("compressedOutput", ctypes.c_bool),
        ("SNRThreshold", ctypes.c_float),
        ("SQIThreshold", ctypes.c_float),
    ]


#
# https://stackoverflow.com/questions/36986929/redirect-print-command-in-python-script-through-tqdm-write
#


class DummyFile(object):
    file = None

    def __init__(self, file):
        self.file = file

    def write(self, x):
        # Avoid print() second call (useless \n)
        if len(x.rstrip()) > 0:
            tqdm.tqdm.write(x, file=self.file)


@contextlib.contextmanager
def nostdout():
    save_stdout = sys.stdout
    sys.stdout = DummyFile(sys.stdout)
    yield
    sys.stdout = save_stdout


def passthrough(x):
    return x


class Workspace(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ("configs", ctypes.POINTER(RKConfig)),
        ("pulses", RKBuffer),
        ("rays", RKBuffer),
        ("fftModule", ctypes.POINTER(RKFFTModule)),
        ("pulseMachine", ctypes.POINTER(RKPulseEngine)),
        ("momentMachine", ctypes.POINTER(RKMomentEngine)),
        ("sweepMachine", ctypes.POINTER(RKSweepEngine)),
        ("recorder", ctypes.POINTER(RKRawDataRecorder)),
        ("ringMachine", ctypes.POINTER(RKPulseRingFilterEngine)),
        ("userModule", RKUserModule),
        ("userModuleFree", ctypes.CFUNCTYPE(UNCHECKED(None), RKUserModule)),
        ("configIndex", pyRKuint32),
        ("pulseIndex", pyRKuint32),
        ("rayIndex", pyRKuint32),
    ]

    def __init__(self):
        super().__init__()
        self.name = f"\033{RKPythonColor[4:]}<  Python Core  >\033[m"
        self.fid = None
        self.desc = None
        self.header = None
        self.filesize = 0
        self.allocated = False
        RKSetProgramName(b"RadarKit")
        RKLog(f"{self.name} Initializing ...")

    def open(self, filename, **kwargs):
        """
        kwargs:
            path: str, default "data"
                The data path
            cores: int, default 4
                The number of cores to use for pulse processing.
            verbose: int, default 0
                The verbosity level.
        """
        self.verbose = kwargs.get("verbose", 0)
        self.fid = RKFileOpen(filename, "r")
        self.header = RKFileHeaderInitFromFid(self.fid).contents
        self.filesize = RKFileGetSize(self.fid)

        if self.verbose:
            RKLog(
                f"{self.name} dataType = {self.header.dataType} out of [{RKRawDataTypeNull} {RKRawDataTypeFromTransceiver} {RKRawDataTypeAfterMatchedFilter}]"
            )

        multiple = RKMemoryAlignSize * self.header.desc.pulseToRayRatio
        newPulseCapacity = (self.header.desc.pulseCapacity + multiple - 1) // multiple * multiple

        # Store as self.desc using the first header.desc and override some attributes. Only the first encounter matters.
        if self.desc is None or self.allocated is False:
            desc = self.header.desc
            desc.dataPath = kwargs.get("path", "data").encode("utf-8")
            desc.configBufferDepth = 3
            desc.pulseBufferDepth = 10 * RKMaximumPulsesPerRay + 50
            desc.rayBufferDepth = RKMaximumRaysPerSweep + 50
            desc.pulseCapacity = newPulseCapacity
            desc.initFlags = (
                RKInitFlagAllocConfigBuffer
                | RKInitFlagAllocRawIQBuffer
                | RKInitFlagAllocMomentBuffer
                | RKInitFlagVerbose
            )
            if self.header.dataType == RKRawDataTypeFromTransceiver:
                useRingFilterEngine = kwargs.get("ring", False)
                if useRingFilterEngine:
                    desc.initFlags |= RKInitFlagStartPulseEngine | RKInitFlagStartRingFilterEngine
                    if self.verbose:
                        print("PulseEngine: Y   RingFilterEngine: Y")
                else:
                    desc.initFlags |= RKInitFlagStartPulseEngine
                    if self.verbose:
                        print("PulseEngine: Y   RingFilterEngine: N")
            elif self.verbose:
                print("PulseEngine: N   RingFilterEngine: N")
            self.configIndex = pyRKuint32(desc.configBufferDepth - 1)
            self.desc = desc
            self.alloc(cores=kwargs.get("cores", 4))

        if (
            (self.desc.initFlags & RKInitFlagStartPulseEngine)
            and (self.header.dataType == RKRawDataTypeAfterMatchedFilter)
        ) or (
            not (self.desc.initFlags & RKInitFlagStartPulseEngine)
            and (self.header.dataType == RKRawDataTypeFromTransceiver)
        ):
            raise RKEngineError("dataType mismatch. Please use a new workspace.")

        k = next_modulo_s(self.configIndex.value, self.desc.configBufferDepth)
        config = self.configs[k]
        ctypes.memmove(ctypes.byref(config), ctypes.byref(self.header.config), ctypes.sizeof(RKConfig))
        config.waveform = self.header.config.waveform
        config.waveformDecimate = self.header.config.waveformDecimate
        self.configIndex.value = k

    def close(self):
        if self.fid is None:
            return
        RKFileClose(self.fid)
        self.fid = None

    def alloc(self, verbose=0, cores=4):
        desc = self.desc

        self.fftModule = RKFFTModuleInit(desc.pulseCapacity, verbose)

        desc.configBufferSize = RKConfigBufferAlloc(ctypes.byref(self.configs), desc.configBufferDepth)
        desc.pulseBufferSize = RKPulseBufferAlloc(ctypes.byref(self.pulses), desc.pulseCapacity, desc.pulseBufferDepth)
        desc.rayBufferSize = RKRayBufferAlloc(
            ctypes.byref(self.rays), desc.pulseCapacity // desc.pulseToRayRatio, desc.rayBufferDepth
        )

        configIndexPtr = ctypes.byref(self.configIndex)
        pulseIndexPtr = ctypes.byref(self.pulseIndex)
        rayIndexPtr = ctypes.byref(self.rayIndex)
        descPtr = ctypes.byref(desc)

        self.pulseMachine = RKPulseEngineInit()
        RKPulseEngineSetVerbose(self.pulseMachine, verbose)
        RKPulseEngineSetEssentials(
            self.pulseMachine, descPtr, self.fftModule, self.configs, configIndexPtr, self.pulses, pulseIndexPtr
        )
        RKPulseEngineSetCoreCount(self.pulseMachine, cores)
        if desc.initFlags & RKInitFlagStartPulseEngine:
            RKPulseEngineStart(self.pulseMachine)

        self.ringMachine = RKPulseRingFilterEngineInit()
        RKPulseRingFilterEngineSetVerbose(self.ringMachine, verbose)
        RKPulseRingFilterEngineSetEssentials(
            self.ringMachine, descPtr, self.configs, configIndexPtr, self.pulses, pulseIndexPtr
        )
        RKPulseRingFilterEngineSetCoreCount(self.ringMachine, 2)
        if desc.initFlags & RKInitFlagStartRingFilterEngine:
            RKPulseRingFilterEngineStart(self.ringMachine)
            RKPulseEngineSetWaitForRingFilter(self.pulseMachine, True)

        self.momentMachine = RKMomentEngineInit()
        RKMomentEngineSetVerbose(self.momentMachine, verbose)
        RKMomentEngineSetEssentials(
            self.momentMachine,
            descPtr,
            self.fftModule,
            self.configs,
            configIndexPtr,
            self.pulses,
            pulseIndexPtr,
            self.rays,
            rayIndexPtr,
        )
        RKMomentEngineSetCoreCount(self.momentMachine, 2)
        RKMomentEngineSetMomentProcessorToMultiLag3(self.momentMachine)
        RKMomentEngineStart(self.momentMachine)

        self.sweepMachine = RKSweepEngineInit()
        RKSweepEngineSetVerbose(self.sweepMachine, verbose)
        RKSweepEngineSetEssentials(
            self.sweepMachine, descPtr, None, self.configs, configIndexPtr, self.rays, rayIndexPtr
        )
        RKSweepEngineSetRecord(self.sweepMachine, True)
        RKSweepEngineStart(self.sweepMachine)

        self.recorder = RKRawDataRecorderInit()
        RKRawDataRecorderSetEssentials(
            self.recorder, descPtr, None, self.configs, configIndexPtr, self.pulses, pulseIndexPtr
        )
        RKRawDataRecorderSetRawDataType(self.recorder, RKRawDataTypeAfterMatchedFilter)
        RKRawDataRecorderSetVerbose(self.recorder, verbose)

        self.allocated = True

        global workspaces
        workspaces.append(self)

    def free(self):
        if self.fid:
            self.close()
        scratch = self.sweepMachine.contents.scratchSpaces[self.sweepMachine.contents.scratchSpaceIndex]
        count = scratch.rayCount + self.sweepMachine.contents.business
        if self.verbose:
            print(f"count = {count}")
        if count > 2:
            RKPulseEngineWaitWhileBusy(self.pulseMachine)
            RKMomentEngineWaitWhileBusy(self.momentMachine)
            RKSweepEngineFlush(self.sweepMachine)
        RKRawDataRecorderSetRecord(self.recorder, False)
        RKReadPulseFromFileReference(None, None, None)

        if (self.userModule is not None) and (self.userModuleFree is not None):
            self.userModuleFree(self.userModule)

        RKPulseEngineStop(self.pulseMachine)
        RKPulseRingFilterEngineStop(self.ringMachine)
        RKMomentEngineStop(self.momentMachine)
        RKSweepEngineStop(self.sweepMachine)
        RKRawDataRecorderStop(self.recorder)

        RKPulseEngineFree(self.pulseMachine)
        RKPulseRingFilterEngineFree(self.ringMachine)
        RKMomentEngineFree(self.momentMachine)
        RKSweepEngineFree(self.sweepMachine)
        RKRawDataRecorderFree(self.recorder)

        RKFFTModuleFree(self.fftModule)
        RKConfigBufferFree(self.configs)
        RKPulseBufferFree(self.pulses)
        RKRayBufferFree(self.rays)

        self.allocated = False

        global workspaces
        workspaces.remove(self)

    def print(self, stuff):
        with nostdout():
            print(stuff)

    def set_waveform(self, samples):
        waveform = RKWaveformInitWithCountAndDepth(1, samples.size)
        waveform.contents.filterAnchors[0][0].origin = 0
        waveform.contents.filterAnchors[0][0].length = samples.size
        waveform.contents.fc = 0
        waveform.contents.fs = self.fs
        waveform.contents.type = RKWaveformTypeIsComplex
        waveform.contents.name = b"Custom"
        waveform.contents.filterCounts[0] = 1
        waveform.contents.filterAnchors[0][0].name = 0
        waveform.contents.filterAnchors[0][0].origin = 0
        waveform.contents.filterAnchors[0][0].length = waveform.contents.depth
        waveform.contents.filterAnchors[0][0].maxDataLength = RKMaximumGateCount
        waveform.contents.filterAnchors[0][0].subCarrierFrequency = 0
        place_RKComplex_array(waveform.contents.samples[0], samples)
        RKWaveformNormalizeNoiseGain(waveform)

        config = self.configs[self.configIndex.value]
        config.waveform = waveform
        config.waveformDecimate = RKWaveformCopy(waveform)
        RKWaveformDecimate(config.waveformDecimate, self.desc.pulseToRayRatio)
        RKPulseEngineSetFilterByWaveform(self.pulseMachine, config.waveform)
        self.blindGateCount = samples.size

    def set_moment_method(self, method):
        if method not in ["pp", "pph", "ppa", "m2", "m3", "m4", "spec"]:
            raise RKEngineError(f"Invalid moment method: {method}")
        if method.startswith("m"):
            self.momentMachine.contents.userLagChoice = int(method[-1])
            method = "mx"
        self.momentMachine.contents.momentProcessor = {
            "pp": ctypes.cast(RKPulsePair, type(self.momentMachine.contents.momentProcessor)),
            "pph": ctypes.cast(RKPulsePairHop, type(self.momentMachine.contents.momentProcessor)),
            "ppa": ctypes.cast(RKPulsePairATSR, type(self.momentMachine.contents.momentProcessor)),
            "mx": ctypes.cast(RKMultiLag, type(self.momentMachine.contents.momentProcessor)),
            "spec": ctypes.cast(RKSpectralMoment, type(self.momentMachine.contents.momentProcessor)),
        }[method]

    def unset_user_module(self):
        if self.userModule is not None:
            self.userModuleFree(self.userModule)
        self.userModule = None
        self.userModuleFree = ctypes.cast(None, type(self.userModuleFree))
        RKPulseEngineUnsetCompressor(self.pulseMachine)
        RKMomentEngineUnsetCalibrator(self.momentMachine)

    def read(self, count=None, filter=None):
        if self.fid is None:
            raise RKEngineError("No file is open.")
        if not self.allocated:
            raise RKEngineError("This workspace has been released.")

        RKRawDataRecorderSetRecord(self.recorder, False)

        config = self.configs[self.configIndex.value]
        RKPulseEngineSetFilterByWaveform(self.pulseMachine, config.waveform)

        pos = RKFileTell(self.fid)

        # Read the first pulse to gather some intels
        pulse = RKPulseEngineGetVacantPulse(self.pulseMachine, RKPulseStatusCompressed)
        r = RKReadPulseFromFileReference(pulse, ctypes.byref(self.header), self.fid)
        if r != RKResultSuccess:
            self.pulseIndex.value = previous_modulo_s(self.pulseIndex.value, self.desc.pulseBufferDepth)
            raise RKEngineError("Failed to read the first pulse.")
        pulse.contents.header.configIndex = self.configIndex.value
        # Estimate the number of pulses
        pulse.contents.header.s |= RKPulseStatusHasIQData | RKPulseStatusHasPosition
        if self.header.dataType == RKRawDataTypeAfterMatchedFilter:
            pulse.contents.header.s |= RKPulseStatusCompleteForMoments
            gateCount = pulse.contents.header.downSampledGateCount
            pulseCount = (self.filesize - pos) // (
                ctypes.sizeof(RKPulseHeader) + 2 * gateCount * ctypes.sizeof(RKComplex)
            )
        else:
            gateCount = pulse.contents.header.gateCount
            pulseCount = (self.filesize - pos) // (
                ctypes.sizeof(RKPulseHeader) + 2 * gateCount * ctypes.sizeof(RKInt16C)
            )
        # Read the first pulse, which should be the same as the one above but use get_done_pulse() for pulseEngine->doneIndex
        pulse = None
        while pulse is None:
            time.sleep(0.05)
            pulse = self.get_done_pulse()
        downSampledGateCount = pulse.contents.header.downSampledGateCount
        print(
            f"Estimated pulseCount = {pulseCount:,d}"
            + f"   gateCount = {gateCount:,d}"
            + f"   downSampledGateCount = {downSampledGateCount:,d}"
        )
        if count is not None:
            pulseCount = min(count, pulseCount)
        # Initialize the arrays and populate with the first pulse
        el = np.zeros((pulseCount,), dtype=np.float32)
        az = np.zeros((pulseCount,), dtype=np.float32)
        if self.header.dataType == RKRawDataTypeFromTransceiver:
            riq = np.zeros((pulseCount, 2, gateCount), dtype=np.complex64)
            riq[0, :, :] = read_RKInt16C_from_pulse(pulse, gateCount)
        else:
            riq = None
        ciq = np.zeros((pulseCount, 2, downSampledGateCount), dtype=np.complex64)
        el[0] = pulse.contents.header.elevationDegrees
        az[0] = pulse.contents.header.azimuthDegrees
        ciq[0, :, :] = read_RKComplex_from_pulse(pulse, downSampledGateCount)

        targetScratchSpaceIndex = self.sweepMachine.contents.lastRecordedScratchSpaceIndex

        with tqdm.tqdm(total=pulseCount, ncols=90, bar_format="{l_bar}{bar}|{elapsed}<{remaining}") as pbar:
            ic = 0
            bz = 1
            for ip in range(1, pulseCount):
                pulse = RKPulseEngineGetVacantPulse(self.pulseMachine, RKPulseStatusCompressed)

                z = 1
                while self.pulseMachine.contents.maxWorkerLag > 0.7 or bz > 0.7 * self.desc.pulseBufferDepth:
                    time.sleep(0.01)
                    s = z * 0.01
                    m = self.pulseMachine.contents.maxWorkerLag
                    # self.print(f"Waiting for workers ...  z = {z:d} / {s:.1f}s   {m:.1f}\n")
                    if z % 100 == 0:
                        s = z * 0.01
                        m = self.pulseMachine.contents.maxWorkerLag
                        print(f"Waiting for workers ...  z = {z:d} / {s:.1f}s   {m:.1f}\n")
                    pulse = self.get_done_pulse()
                    while pulse is not None:
                        ic += 1
                        bz -= 1
                        ciq[ic, :, :] = read_RKComplex_from_pulse(pulse, downSampledGateCount)
                        pulse = self.get_done_pulse()
                    z = z + 1

                r = RKReadPulseFromFileReference(pulse, ctypes.byref(self.header), self.fid)
                if r != RKResultSuccess:
                    self.pulseIndex.value = previous_modulo_s(self.pulseIndex.value, self.desc.pulseBufferDepth)
                    self.print("No more data to read.")
                    ip -= 1
                    break
                pulse.contents.header.configIndex = self.configIndex.value
                # Insert pre-filter here
                if filter is not None:
                    d16 = read_RKInt16C_from_pulse(pulse, pulse.contents.header.gateCount)
                    place_RKInt16C_to_pulse(pulse, filter(d16))

                el[ip] = pulse.contents.header.elevationDegrees
                az[ip] = pulse.contents.header.azimuthDegrees
                if self.header.dataType == RKRawDataTypeFromTransceiver:
                    riq[ip, :, :] = read_RKInt16C_from_pulse(pulse, gateCount)

                pulse.contents.header.s |= RKPulseStatusHasIQData | RKPulseStatusHasPosition
                if not (self.desc.initFlags & RKInitFlagStartRingFilterEngine):
                    pulse.contents.header.s |= RKPulseStatusRingProcessed
                if self.header.dataType == RKRawDataTypeAfterMatchedFilter:
                    pulse.contents.header.s |= RKPulseStatusCompleteForMoments

                bz += 1

                pulse = self.get_done_pulse()
                while pulse is not None:
                    ic += 1
                    bz -= 1
                    ciq[ic, :, :] = read_RKComplex_from_pulse(pulse, downSampledGateCount)
                    pulse = self.get_done_pulse()

                # Yield to other threads to process the pulses. RKPulseEngineUpdateMinMaxWorkerLag() could take up to 50us
                time.sleep(50.0e-6)

                pbar.update(1.0)
            pbar.close()

        if self.verbose > 1:
            print(f"E1: ip = {ip:,d}   ic = {ic:,d}   pulseCount = {pulseCount:,d}")
        s = 0
        while ic < ip and s < 30:
            pulse = self.get_done_pulse()
            if pulse is None:
                time.sleep(0.1)
                s += 1
                continue
            ic += 1
            ciq[ic, :, :] = read_RKComplex_from_pulse(pulse, downSampledGateCount)
        if self.verbose > 1 or s >= 30 or ic < ip:
            print(f"E0: ip = {ip:,d}   ic = {ic:,d}   pulseCount = {pulseCount:,d}")

        RKFileSeek(self.fid, pos)
        RKPulseEngineWaitWhileBusy(self.pulseMachine)
        RKMomentEngineWaitWhileBusy(self.momentMachine)

        s = 0
        while self.sweepMachine.contents.lastRecordedScratchSpaceIndex == targetScratchSpaceIndex and s < 50:
            time.sleep(0.1)
            s += 1

        return {"riq": riq, "ciq": ciq, "el": el, "az": az}

    def get_current_config(self):
        return self.configs[self.configIndex.value]

    def get_done_pulse(self):
        pulse = RKPulseEngineGetProcessedPulse(self.pulseMachine, False)
        return pulse if pulse else None

    def get_moment(self, offset=None, variable_list=["S", "Z", "V", "W", "D", "R", "P"]):
        if offset is None:
            if self.verbose:
                print(
                    f"rayIndex = {self.rayIndex.value}"
                    + f"   sweepMachine.business = {self.sweepMachine.contents.business}"
                    + f"   momentMachine.business = {self.momentMachine.contents.business}"
                )
            k = self.sweepMachine.contents.lastRecordedScratchSpaceIndex
        else:
            k = offset
        scratch = self.sweepMachine.contents.scratchSpaces[k]
        if self.verbose:
            print(f"Gathering {scratch.rayCount} rays from scratch space #{k} ...")
        self.variables = {}
        gateCount = scratch.rays[0].contents.header.gateCount
        scanTime = scratch.rays[0].contents.header.startTime.tv_sec
        elevations = np.array([scratch.rays[k].contents.header.startElevation for k in range(scratch.rayCount)])
        azimuths = np.array([scratch.rays[k].contents.header.startAzimuth for k in range(scratch.rayCount)])
        ranges = np.arange(scratch.rays[0].contents.header.gateCount) * scratch.rays[0].contents.header.gateSizeMeters
        for varname in variable_list:
            buf = []
            for k in range(scratch.rayCount):
                ray = scratch.rays[k]
                data = read_RKFloat_array(RKGetFloatDataFromRay(ray, Productdict[varname]), gateCount)
                if data is None or data.size == 0:
                    print(f"Breaking away at k = {k} ...")
                    break
                else:
                    buf.append(data)
            self.variables.update({varname: np.asarray(buf)})
        self.variables.update({"time": scanTime})
        self.variables.update({"elevations": elevations})
        self.variables.update({"azimuths": azimuths})
        self.variables.update({"ranges": ranges})
        # self.variables.update({"elevations": np.asarray([ray.contents.header.startElevation for ray in scratch.rays])})
        if "P" in self.variables:
            self.variables["P"] = np.degrees(self.variables["P"])
        return self.variables

    def flush(self):
        RKSweepEngineFlush(self.sweepMachine)


def open(filename, **kwargs):
    global workspaces
    force = kwargs.get("force", False)
    if force or len(workspaces) == 0:
        workspace = Workspace()
    else:
        workspace = workspaces[-1]
    workspace.open(filename, **kwargs)
    if len(workspaces) > 3:
        print(f"Warning. Multiple workspaces ({len(workspaces)}) detected.")
    return workspace


def close():
    global workspaces
    for workspace in workspaces:
        workspace.free()
    workspaces = []


def last():
    global workspaces
    if len(workspaces) == 0:
        raise RKEngineError("No workspace is open.")
    return workspaces[-1]


def previous_modulo_s(i, S):
    return (S - 1) if i == 0 else (i - 1)


def next_modulo_s(i, S):
    return 0 if i == S - 1 else i + 1


def read_RKFloat_array(rkpos, count):
    return np.ctypeslib.as_array(ctypes.cast(rkpos, ctypes.POINTER(RKFloat)), (count,))


def read_RKInt16C_array(rkpos, count):
    bufiq = np.ctypeslib.as_array(ctypes.cast(rkpos, ctypes.POINTER(ctypes.c_int16)), (count, 2))
    return bufiq[:, 0] + 1j * bufiq[:, 1]


def read_RKComplex_array(rkpos, count):
    bufiq = np.ctypeslib.as_array(ctypes.cast(rkpos, ctypes.POINTER(ctypes.c_float)), (count, 2))
    return bufiq[:, 0] + 1j * bufiq[:, 1]


def read_RKInt16C_from_pulse(pulse, count):
    h = ctypes.cast(RKGetInt16CDataFromPulse(pulse, 0), ctypes.POINTER(ctypes.c_int16))
    h = np.ctypeslib.as_array(h, (count * 2,)).astype(np.float32).view(np.complex64)
    v = ctypes.cast(RKGetInt16CDataFromPulse(pulse, 1), ctypes.POINTER(ctypes.c_int16))
    v = np.ctypeslib.as_array(v, (count * 2,)).astype(np.float32).view(np.complex64)
    x = np.vstack((h, v))
    np.conj(x, out=x)
    return x


def read_RKComplex_from_pulse_v1(pulse, count):
    h = ctypes.cast(RKGetComplexDataFromPulse(pulse, 0), ctypes.POINTER(ctypes.c_float))
    h = np.ctypeslib.as_array(h, (count * 2,)).view(np.complex64)
    v = ctypes.cast(RKGetComplexDataFromPulse(pulse, 1), ctypes.POINTER(ctypes.c_float))
    v = np.ctypeslib.as_array(v, (count * 2,)).view(np.complex64)
    x = np.vstack((h, v))
    np.conj(x, out=x)
    return x


def read_RKComplex_from_pulse(pulse, count):
    _from_mem = ctypes.pythonapi.PyMemoryView_FromMemory
    _from_mem.restype = ctypes.py_object
    h = np.frombuffer(_from_mem(RKGetComplexDataFromPulse(pulse, 0), 8 * count), dtype=np.complex64)
    v = np.frombuffer(_from_mem(RKGetComplexDataFromPulse(pulse, 1), 8 * count), dtype=np.complex64)
    x = np.vstack((h, v))
    np.conj(x, out=x)
    return x


def place_RKComplex_array(dst, src):
    np.conj(src, out=src)
    ctypes.memmove(ctypes.cast(dst, ctypes.POINTER(ctypes.c_float)), src.flatten().ctypes.data, src.nbytes)


def place_RKInt16C_array_v1(dst, src):
    buf = np.zeros((src.size * 2), dtype=np.int16)
    buf[::2] = src.real.astype(np.int16)
    buf[1::2] = -src.imag.astype(np.int16)
    ctypes.memmove(ctypes.cast(dst, ctypes.POINTER(ctypes.c_int16)), buf.ctypes.data, buf.nbytes)


def place_RKInt16C_to_pulse(pulse, src):
    buf = np.zeros((src.shape[1] * 2), dtype=np.int16)
    for p in range(2):
        dst = RKGetInt16CDataFromPulse(pulse, p)
        buf[::2] = src.real[p, :].astype(np.int16)
        buf[1::2] = -src.imag[p, :].astype(np.int16)
        ctypes.memmove(ctypes.cast(dst, ctypes.POINTER(ctypes.c_int16)), buf.ctypes.data, buf.nbytes)


def read_RKComplex_from_waveform(waveform, index=0):
    _from_mem = ctypes.pythonapi.PyMemoryView_FromMemory
    _from_mem.restype = ctypes.py_object
    return np.frombuffer(_from_mem(waveform.contents.samples[index], 8 * waveform.contents.depth), dtype=np.complex64)
